# -*- coding: utf-8 -*-
#
# Copyright © 2008  Red Hat, Inc. All rights reserved.
# Copyright © 2009  Mandriva. All rights reserved.
# Copyright © 2009  Jolicloud. All rights reserved.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. Any Red Hat trademarks that are
# incorporated in the source code or documentation are not subject to the GNU
# General Public License and may only be used or replicated with the express
# permission of Red Hat, Inc.
#
# Author(s): Luke Macken <lmacken@redhat.com>
#            Aurelien Lefebvre <alefebvre@mandriva.com>
#            Benjamin Ryzman <benjamin@jolicloud.com>

import subprocess
import logging
import os
import string
import commands
import re
import signal

from datetime import datetime
from stat import ST_SIZE
from threading import Thread
from time import sleep

from liveusb import _

class LiveUSBError(Exception):
    def __init__(self, value): self.value = value
    def __str__(self): return str(self.value)

class LiveUSBCreator(object):
    """ An OS-independent parent class for Live USB Creators """

    iso = None
    drives = {}
    pids = []
    isosize = 0
    log = None
    drive = None

    def __init__(self, opts):
        self.opts = opts
        self._setup_logger()

    def _setup_logger(self):
        self.log = logging.getLogger(__name__)
        level = logging.INFO
        if self.opts.verbose:
            level = logging.DEBUG
        self.log.setLevel(level)
        self.handler = logging.StreamHandler()
        self.handler.setLevel(level)
        formatter = logging.Formatter("[%(module)s:%(lineno)s] %(message)s")
        self.handler.setFormatter(formatter)
        self.log.addHandler(self.handler)

    def detect_removable_drives(self):
        """ This method should populate self.drives with removable devices """
        raise NotImplementedError

    def terminate(self):
        """ Terminate any subprocesses that we have spawned """
        raise NotImplementedError

    def set_iso(self, iso):
        """ Select the given ISO """
        self.iso = self._to_unicode(iso)
        self.isosize = os.stat(self.iso)[ST_SIZE]

    def _to_unicode(self, obj, encoding='utf-8'):
        if hasattr(obj, 'toUtf8'): # PyQt4.QtCore.QString
            obj = unicode(obj.toUtf8(), "utf-8")
        #if isinstance(obj, basestring):
        #    if not isinstance(obj, unicode):
        #        obj = unicode(obj, encoding, 'replace')
        return obj

    def _test_hybrid_1(self, iso):
        hybrid = False
        for i in range(0,512):
            if not iso.read(1) == '\x00':
                hybrid = True
            if hybrid:
                break
        return hybrid

    def _test_hybrid_2(self, iso):
        hybrid = False
        iso.seek(0x1fe)
        if iso.read(1) == '\x55':
            iso.seek(0x1ff)
            if iso.read(1) == '\xaa':
                hybrid = True
        return hybrid

    def _test_hybrid_3(self, iso):
        hybrid = True
        iso.seek(0x200)
        for i in range(0x200,0x8000):
            if iso.read(1) != '\x00':
                hybrid = False
                break
        return hybrid

    def is_hybrid(self, iso):
        isofile = open(iso, "rb")
        hybrid =  self._test_hybrid_1(isofile) and self._test_hybrid_2(isofile) and self._test_hybrid_3(isofile)
        isofile.close()
        return hybrid

    def get_device(self):
        drive = self._to_unicode(self.drive)
        device = None
        if self.drives.has_key(drive):
            device = self.drives[drive]['device']
        return device

    def build_disk(self, progress_thread, hfile, hdevice, sectorsize = 512):
        from hashlib import md5

        self.log.info(_("Writing ISO on USB..."))
        sect = 0
        filemd5 = md5()
        required_sectors = os.path.getsize(self.iso) / sectorsize
        while sect < required_sectors:
            os.lseek(hfile, sect*sectorsize, os.SEEK_SET)
            os.lseek(hdevice, sect*sectorsize, os.SEEK_SET)
            count = required_sectors - sect
            if count > 16384:
                count = 16384 # 8M
            data = os.read(hfile, sectorsize*count)
            filemd5.update(data)
            os.write(hdevice, data)
            progress = (float(sect)/required_sectors)*100
            progress = round(progress, 2)
            progress_thread.update(progress)
            sect += count

        return filemd5.digest()

    def check_disk(self, progress_thread, hdevice, digest, sectorsize = 512):
        from hashlib import md5

        self.log.info(_("Checking disk content..."))
        sect = 0
        devicemd5 = md5()
        required_sectors = os.path.getsize(self.iso) / sectorsize
        while sect < required_sectors:
            os.lseek(hdevice, sect*sectorsize, os.SEEK_SET)
            count = required_sectors - sect
            if count > 16384:
                count = 16384
            data = os.read(hdevice, sectorsize*count)
            devicemd5.update(data)
            progress = (float(sect)/required_sectors)*100
            progress = round(progress, 2)
            progress_thread.update(progress)
            sect += count

        if digest != devicemd5.digest():
            raise LiveUSBError(_("Disk content corrupted!"))
class UNIXLiveUSBCreator(LiveUSBCreator):

    def umount_device(self):
        raise NotImplementedError

    def terminate(self):
        self.log.info("Cleaning up...")
        for pid in self.pids:
            try:
                os.kill(pid, signal.SIGHUP)
                self.log.debug("Killed process %d" % pid)
            except OSError:
                pass

    def build_disk(self, progress_thread):
        if os.getuid() != 0:
            raise LiveUSBError(_("You must run this application as an administrator"))

        isosize = float(self.isosize)
        self.umount_device()

        hfile = os.open(self.iso, os.O_RDONLY | os.O_EXCL) # Unbuffered
        hdevice = os.open(self.get_device(), os.O_WRONLY | os.O_EXCL | os.O_SYNC)
        digest = LiveUSBCreator.build_disk(self, progress_thread, hfile, hdevice)
        os.close(hfile)
        os.close(hdevice)

        hdevice = os.open(self.get_device(), os.O_RDONLY)
        LiveUSBCreator.check_disk(self, progress_thread, hdevice, digest)
        os.close(hdevice)

class LinuxLiveUSBCreator(UNIXLiveUSBCreator):

    bus = None # the dbus.SystemBus
    hal = None # the org.freedesktop.Hal.Manager dbus.Interface

    def umount_device(self):
        # we need to umount any mounted partition on the device first
        # FIXME do it using HAL
        umount_cmd = '/usr/sbin/umount.hal'
        if not (os.path.exists(umount_cmd) and os.access(umount_cmd, os.X_OK)):
            umount_cmd = '/bin/umount'
        p = re.compile(self.get_device()+"\w*")
        umount = p.findall(subprocess.Popen('/bin/mount',
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env={"LC_ALL": "C"}).communicate()[0])
        if len(umount):
            umount.insert(0,umount_cmd)
            ret = subprocess.call(umount)
            if ret != 0:
                raise LiveUSBError(_("/bin/umount command failed"))
            self.log.info(_("%s unmounted" %(self.get_device())))

    def detect_removable_drives(self):
        """ Detect all removable USB storage devices using HAL via D-Bus """
        import dbus
        self.drives = {}
        self.bus = dbus.SystemBus()
        hal_obj = self.bus.get_object("org.freedesktop.Hal",
                "/org/freedesktop/Hal/Manager")
        self.hal = dbus.Interface(hal_obj, "org.freedesktop.Hal.Manager")

        devices = []
        devices = self.hal.FindDeviceByCapability("storage")

        for udi in devices:
            try:
                dev = self._get_device_from_udi(udi)
                if dev.GetProperty("storage.bus") == "usb":
                    self._add_device(dev)
            except dbus.exceptions.DBusException:
                pass

        if not len(self.drives):
            raise LiveUSBError(_("Unable to find any USB drives"))

    def _add_device(self, dev):
        device = None
        model = ''
        vendor = ''
        capacity = 0
        try:
            self.log.debug('dev: %s' % (dev))
            device = dev.GetProperty('block.device')
            self.log.debug('device: %s' % (device))
            capacity = str(dev.GetProperty('storage.removable.media_size'))
            self.log.debug('capacity: %s' % (capacity))
            vendor = dev.GetProperty('storage.vendor')
            self.log.debug('vendor: %s' % (vendor))
            model = dev.GetProperty('storage.model')
            self.log.debug('model: %s' % (model))
        except Exception:
            pass

        if device and capacity:
            self.drives[device] = {
                    'name'      : device,
                    'device'    : device,
                    'model'     : vendor + ' ' + model,
                    'capacity'  : capacity
                    }

    def _get_device_from_udi(self, udi):
        """ Return a dbus Interface to a specific HAL device UDI """
        import dbus
        dev_obj = self.bus.get_object("org.freedesktop.Hal", udi)
        return dbus.Interface(dev_obj, "org.freedesktop.Hal.Device")

class WindowsLiveUSBCreator(LiveUSBCreator):

    def get_device_id(self):
        drive = self._to_unicode(self.drive)
        id = None
        if self.drives.has_key(drive):
            id = self.drives[drive]['id']
        return id

    def detect_removable_drives(self):
        import win32file
        import win32api
        from win32com.shell import shell
        from liveusb.win32utils import _getPhysicalDeviceId
        import pywintypes
        if not shell.IsUserAnAdmin():
            raise LiveUSBError(_("You must run this application as an administrator"))


        avail_drives = win32api.GetLogicalDrives()
        i = 0
        while avail_drives:
            if avail_drives % 2:
                letter = chr(ord('A') + i)
                name = letter + ':'
                device = '\\\\.\\' + name
                capacity = '!'
                if win32file.DRIVE_REMOVABLE == win32file.GetDriveType(name):
                    try:
                        capacity = win32api.GetDiskFreeSpaceEx(name)[1]
                    except pywintypes.error, e:
                        if e.winerror == 1005: # Unknown filesystem
                            pass
                    self.drives[device] = {
                            'name'      : name,
                            'device'    : device,
                            'id'        : _getPhysicalDeviceId(device),
                            'capacity'  : capacity
                            }
            i += 1
            avail_drives >>= 1

    def build_disk(self, progress_thread):
        from win32con import GENERIC_WRITE, GENERIC_READ
        from liveusb.win32utils import _getHandleOnVolume, _getLockOnVolumeByHandle, _unmountVolumeByHandle, _getHandleOnDevice, _getNumberOfSectorsAndSectorSizeByHandle, _removeLockOnVolumeByHandle

        hvolume = _getHandleOnVolume(self.get_device(), GENERIC_WRITE)
        _getLockOnVolumeByHandle(hvolume)
        _unmountVolumeByHandle(hvolume)

        hdevice = _getHandleOnDevice(self.get_device_id(), GENERIC_WRITE | GENERIC_READ)
        (available_sectors, sectorsize) = _getNumberOfSectorsAndSectorSizeByHandle(hdevice)
        hdevice.Close()

        hfile = os.open(self.iso, os.O_RDONLY | os.O_EXCL | os.O_BINARY) # Unbuffered
        hdevice = os.open("\\\\.\\PhysicalDrive%d" % self.get_device_id(), os.O_WRONLY | os.O_EXCL | os.O_BINARY)
        digest = LiveUSBCreator.build_disk(self, progress_thread, hfile, hdevice, sectorsize)
        os.close(hfile)
        os.close(hdevice)

        hdevice = os.open("\\\\.\\PhysicalDrive%d" % self.get_device_id(), os.O_RDONLY | os.O_BINARY)
        LiveUSBCreator.check_disk(self, progress_thread, hdevice, digest, sectorsize)
        os.close(hdevice)

        _removeLockOnVolumeByHandle(hvolume)
        hvolume.close()
        
    def terminate(self):
        pass

class DarwinLiveUSBCreator(UNIXLiveUSBCreator):
    def umount_device(self):
        # convert /dev/rdisk# back to /dev/disk#
        pattern = "/dev/rdisk(\d+)"
        drive = re.compile(pattern).sub(r'/dev/disk\1',self.get_device())
        [status, out] = commands.getstatusoutput('/usr/sbin/diskutil unmountDisk "%s"' %(drive) )
        self.log.debug(out)
        # there is a regression in Snow Leopard where diskutil unmountDisk always returns 1
        # if status != 0:
        pattern = "Unmount of all volumes on .* was successful"
        if not re.compile(pattern).match(out):
                raise LiveUSBError(_("Error while umounting device"))
        self.log.info(_("%s unmounted" %(self.get_device())))

    def detect_removable_drives(self):
        """ Detect all removable USB storage devices using DiskUtil"""
        self.drives = {}
        [status, out] = commands.getstatusoutput('/usr/sbin/diskutil list | grep ^/dev')
        drives = out.split("\n")
        for drive in drives:
            #print drive
            Protocol = None
            Ejectable = "No"
            Model = "Unknown"
            name = drive
            capacity = 0
            [status, out] = commands.getstatusoutput('/usr/sbin/diskutil info "%s"' %(drive) )
            driveinfo = out.split("\n")
            for info in driveinfo:
                #print info
                words = string.split(info, ':')
                #print words
                if words[0].strip() == "Protocol":
                    Protocol = words[1].strip()
                if words[0].strip() == "Ejectable":
                    Ejectable = words[1].strip()
                if words[0].strip() == "Device / Media Name":
                    Model = words[1].strip()
                if words[0].strip() == "Total Size":
                    p = "\((\d+) B\)"
                    match = re.compile(p).search(words[1])
                    if match:
                        capacity = match.group(1)
                if words[0].strip() == "Mount Point":
                    p = "\/Volumes\/(.*)"
                    match = re.compile(p).search(words[1])
                    if match:
                        name = match.group(1)
            if Protocol == "USB" and Ejectable == "Yes":
                # convert /dev/disk# to /dev/rdisk#
                diskPattern = "/dev/disk(\d+)"
                diskPatternCompiled = re.compile(diskPattern)
                drive = diskPatternCompiled.sub(r'/dev/rdisk\1',drive)
                label = None
                self.drives[drive] = {
                        'device'    : drive,
                        'name'      : name,
                        'capacity'  : capacity,
                        'model'     : Model,
                        }
                if not len(self.drives):
                    raise LiveUSBError(_("Unable to find any removable devices"))
